from itertools import product
import numpy
import xarray as xr

from datacube.virtual import Transformation, Measurement
from datacube.storage.masking import valid_data_mask
from fc import __version__
from fc.fractional_cover import fractional_cover

FC_MEASUREMENTS = [
    {
        'name': 'pv',
        'dtype': 'int8',
        'nodata': -1,
        'units': 'percent'
    },
    {
        'name': 'npv',
        'dtype': 'int8',
        'nodata': -1,
        'units': 'percent'
    },
    {
        'name': 'bs',
        'dtype': 'int8',
        'nodata': -1,
        'units': 'percent'
    },
    {
        'name': 'ue',
        'dtype': 'int8',
        'nodata': -1,
        'units': ''
    },
]


class FractionalCover(Transformation):
    """ Applies the fractional cover algorithm to surface reflectance data.
    Requires bands named 'green', 'red', 'nir', 'swir1', 'swir2'
    """

    def __init__(self, regression_coefficients=None, c2_scaling=False):
        if regression_coefficients is None:
            regression_coefficients = {band: [0, 1]
                                       for band in ['green', 'red', 'nir', 'swir1', 'swir2']
                                       }
        self.regression_coefficients = regression_coefficients
        self.c2_scaling = c2_scaling

    def measurements(self, input_measurements):
        return {m['name']: Measurement(**m) for m in FC_MEASUREMENTS}

    def compute(self, data):
        # Typically creates a list of dictionaries looking like [{time: 1234}, {time: 1235}, ...]
        if self.c2_scaling:
            # The C2 data need to be scaled
            for band in ['green', 'red', 'nir', 'swir1', 'swir2']:
                dtype = data[band].dtype
                nodata = data[band].attrs['nodata']

                is_valid_array = valid_data_mask(data[band])

                data[band] = data[band].where(is_valid_array)
                data[band] = numpy.clip((data[band] * 2.75e-5 - 0.2) * 10000, 0, 10000)
                data[band] = data[band].astype(dtype).where(is_valid_array, nodata)
                

        sel = [dict(p)
               for p in product(*[[(i.name, i.item()) for i in c]
                                  for v, c in data.coords.items()
                                  if v not in data.geobox.dims])]
        fc = []
        measurements = [Measurement(**m) for m in FC_MEASUREMENTS]
        for s in sel:
            fc.append(fractional_cover(data.sel(**s), measurements, self.regression_coefficients))
        fc = xr.concat(fc, dim='time')
        fc.attrs['crs'] = data.attrs['crs']
        try:
            fc.rename(inplace=True, BS='bs', PV='pv', NPV='npv', UE='ue')
        except ValueError:   # Assuming the names are already correct and don't need to be changed.
            pass
        return fc

    def algorithm_metadata(self):
        return {
            'algorithm': {
                'name': 'Fractional Cover',
                'version': __version__,
                'repo_url': 'https://github.com/GeoscienceAustralia/fc.git',
                'parameters': {'regression_coefficients': self.regression_coefficients}
            }}


class FakeFractionalCover(Transformation):
    """ Applies the fractional cover algorithm to surface reflectance data.
    Requires bands named 'green', 'red', 'nir', 'swir1', 'swir2'
    """
    def __init__(self, regression_coefficients=None, c2_scaling=False):
        if regression_coefficients is None:
            regression_coefficients = {band: [0, 1]
                                       for band in ['green', 'red', 'nir', 'swir1', 'swir2']
                                       }
        self.regression_coefficients = regression_coefficients
        self.c2_scaling = c2_scaling

    def measurements(self, input_measurements):
        return self.output_measurements

    def compute(self, data):
        if self.c2_scaling:
            # The C2 data need to be scaled
            for band in ['blue', 'green', 'red', 'nir', 'swir1', 'swir2']:
                dtype = data[band].dtype
                nodata = data[band].attrs['nodata']

                is_valid_array = valid_data_mask(data[band])

                data[band] = data[band].where(is_valid_array)
                data[band] = numpy.clip((data[band] * 2.75e-5 - 0.2) * 10000, 0, 10000)
                data[band] = data[band].astype(dtype).where(is_valid_array, nodata)
        return xr.Dataset({'blue': data.blue,
                            'red': data.red,
                           'green': data.green},
                          attrs=data.attrs)
